package xmlutil;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Stack;

/** <p>XMLUtil is a SAX events handler which understands a simple object
    construction vocabulary designed for extracing patterns like lists
    and hashes from an XML document; it can also allow custom hooks.
    The intention is that you will use a tool like XSLT to transform
    the XML you want to parse into the appropriate form for XMLUtil,
    which can then handle the boring and repetitive task of parsing the
    SAX events.</p>

    <p>XMLUtil's XML vocabulary consists of a set of content constructors
    which may be inserted into each other etc etc.  Each content
    constructor produces one java object.</p>

    <p>There are two attributes which may appear on any content constructor.
    <ul>
    <li><b>id</b>: the id attribute specifies a globally unique key which
    can be referenced once construction is complete.
    <li><p><b>custom</b>: the custom attribute indicates that a custom handler
    should be used to create the final Object for this content constructor.
    In this case, once the content constructor is finished constructing as it
    normally would, the handler named in the custom attribute will be passed
    the constructed object and will be expected to return an Object to be 
    constructed in its stead.  So, for example, one might do:</p>
    <pre>
    <hash custom="address">
    <item key="Street"><string>Verndale St.</string></item>
    <item key="town"><string>Brookline</string></item>
    <item key="state"><string>MA</string></item>
    </hash>
    </pre>
    <p>This would cause a hashtable to be initially constructed with the keys
    "street", "town", and "state".  The custom handler object registered under
    the name "address" would then be invoked with this hashtable and would
    return a custom object, likely some representation of an address, which 
    would be inserted.</p>
    </ul>

    <p>The different type of content constructors:</p>

    <h3>Hashtables</h3>
    <pre>
    XML Format:
    &lt;hash [id="yyy"]>
    &lt;item>
    &key>&lt;!-- content constructor here -->&lt;/key> 
	&lt;value>&lt;!-- content constructor here -->&lt;/value>
    &lt;/item>
    <!-- a shorthand form of item that may be used is: -->
    <item key="xxx"> <!-- uses the string 'xxx' as the key -->
    <!-- content constructor here -->
    </item>
    ...
    &lt;/hash>
    </pre>
    <p>Once the document is parsed, a hash element will result in the
    creation of a java.util.Hashtable, with the values and keys one would
    expect.   The shorter form of the 'item' element automatically constructs
    a String object with the value of the 'key' attribute as the key.  The
    longer form can be used if non String elements are desired or simply to be
    pedantic.  They are totaly equivalent.</p>

    <h3>Vectors</h3>
    <pre>
    XML Format:
    &lt;vect [id="xxx"]>
    &lt;!-- content constructors go here -->
    &lt;/vect>
    </pre>
    <p>vect elements will result in the creation of a java.util.Vector,
    with the list of elements one would expect.</p>

    <h3>Strings</h3>
    <pre>
    XML Format:
    &lt;string>string text goes here&lt;/string>
    </pre>
    <p>String elements result in the creation of a java.lang.String object.</p>

    <h3>Longs</h3>
    <pre>
    XML Format:
    &lt;long>integer text goes here&lt;/long>
    </pre>
    <p>Long elements result in a java.lang.Long object</p>

    <h3>References</h3>
    <pre>
    XML Format:
    &lt;ref>xxx&lt;/ref>
    </pre>
    <p>Uses the same object that was previously created with the id 'xxx'</p>
*/
public class XMLUtil
    extends org.xml.sax.helpers.DefaultHandler
{
    public XMLUtil ()
    {
        mIDs = new Hashtable ();
        mCustoms = new Hashtable ();
        mContent = new Stack ();
        mContent.push (new InProgress (this));
    }

    /** each content constructor with an id is placed in here */
    protected Hashtable mIDs;

    /** custom handler objects are placed in here */
    protected Hashtable mCustoms;

    /** adds a custom handler that may then be referenced in the constructed
        document */
    public void registerCustomHandler (String s, Handler h)
    {
        mCustoms.put (s, h);
    }

    /** look up a handler from those registered */
    public Handler getHandler (String s)
    {
        return ((Handler)mCustoms.get (s));
    }

    /** after parsing a document, use this to retrieve constructor objects
        by id */
    public Object get (String id)
    {
        return mIDs.get (id);
    }

    /** during parsing, this is called to register objects with a particular
        id.  You can also call it before parsing, in which case the objects
        can be referenced by a '<ref/>' during parsing. */
    public void put (String id, Object o)
    {
        mIDs.put (id, o);
    }

    /** after parsing a document, use this to retrieve the top level result
     */
    public Object getResult ()
    {
        return mResult;
    }
    
    /** convenience routine to do the cast to Hashtable for you */
    public Hashtable getResultAsHash ()
    {
        return (Hashtable)mResult;
    }

    /** convenience routine to do the cast to Hashtable for you */
    public Vector getResultAsVect ()
    {
        return (Vector)mResult;
    }

    // Parsing State

    protected class InProgress
    {
        public InProgress (String val)
        {
            mID = null;
            mCustom = null;
            mObj = val;
            mType = STRING;
        }

        public InProgress (XMLUtil u)
        {
            mID = null;
            mCustom = null;
            mObj = u;
            mType = TOP;
        }

        public InProgress (String id, String custom, int type) 
	    throws XMLUtilException
        {
            mID = id;
            mCustom = custom;
            mType = type;
            switch (type) {
              case TOP: throw new XMLUtilException ("Bad constructor");
              case HASH:		  mObj = new Hashtable (); break;
              case HASHITEM:      mObj = new Object[2]; break;
              case HASHITEMKEY:
              case HASHITEMVAL:   mObj = null; break;
              case VECT:          mObj = new Vector (); break;
              case LONG:  
              case STRING:        mObj = new StringBuffer (); break;
              case REFERENCE:     mObj = new StringBuffer (); break;
              default:
                throw new XMLUtilException ("bad type" + type);
            }
        }

        public void filter (XMLUtil xu) throws XMLUtilException
        {
            if (mType == STRING) mObj = mObj.toString();
            if (mType == LONG) mObj = Long.valueOf (mObj.toString());
            if (mType == REFERENCE) {
                mObj = xu.get (mObj.toString());
                if (mObj == null) {
                    throw new XMLUtilException ("No object w/ id '" +
                                                mObj + "' could be found.");
                }
            }

            // Did they have a custom handler?
            if (mCustom != null) {
                Handler h = xu.getHandler (mCustom);
                if (h == null) 
                  throw new XMLUtilException ("No custom handler " +
                                              mCustom + " declared.");
                mObj = h.translate (mID, mObj);
            }

            // Did they specify an ID?  If so, enter them into the
            // global hash
            if (mID != null)
              xu.put (mID, mObj);
        }

        /** If this is a type of object that cares about char data, then 
            it processes it.  Otherwise, the char data is ignored unless
            it is not white space, in which case an exception is thrown. */
        public void characters (char[] ch, int start, int length)
	    throws XMLUtilException
        {
            String s = new String (ch, start, length);
            switch (mType) {
              case LONG:  
              case STRING:
              case REFERENCE:
                ((StringBuffer)mObj).append (s);
                return;
              case HASH:
              case VECT:
              case HASHITEM:
              case HASHITEMKEY:
              case HASHITEMVAL:
              case TOP:
                if (s.trim().length() != 0)
                  throw new XMLUtilException ("Invalid character data: "+s);
                return;
              default:
                throw new XMLUtilException ("bad type" + mType + ": " + s);
            }
        }

        /** when a object constructor finishes, its parent's "receive" method
            is called */
        public void receive (InProgress obj) 
	    throws XMLUtilException
        {
            switch (mType) {
              case TOP:		((XMLUtil)mObj).receive (obj); break;
              case HASH:		((Hashtable)mObj).put(obj.getKey(), 
                                                      obj.getVal()); break;
              case HASHITEM:      
                if (obj.mType == HASHITEMKEY)
                  ((Object[])mObj)[0] = obj.mObj; 
                else if (obj.mType == HASHITEMVAL)
                  ((Object[])mObj)[1] = obj.mObj; 
                else 
                  throw new XMLUtilException 
                    ("Invalid mType received by HASHITEM");
                break;
              case VECT:       ((Vector)mObj).addElement (obj.mObj); break;
              case LONG:       throw new XMLUtilException("Long in receive");
              case STRING:     throw new XMLUtilException("String in receive");
              case REFERENCE:  throw new XMLUtilException("Ref in receive");
              case HASHITEMVAL:   
              case HASHITEMKEY:mObj = obj.mObj; break;
              default:
                throw new XMLUtilException ("Invalid mType in receive");
            }
        }

        /** If this is HASHITEM, returns the key */
        public Object getKey() throws XMLUtilException {
            if (mType != HASHITEM) 
              throw new XMLUtilException 
                ("InProgress object is not HASHITEM");
            return ((Object[])mObj)[0];
        }

        /** If this is HASHITEM, returns the value */
        public Object getVal() throws XMLUtilException {
            if (mType != HASHITEM) 
              throw new XMLUtilException 
                ("InProgress object is not HASHITEM");
            return ((Object[])mObj)[1];
        }

        /** ID attribute specified, if any */
        public String mID;

        /** Custom filter specified, if any */
        public String mCustom;

        /** object being constructed */
        public Object mObj;

        /** type of object being constructed, one of the constants below.
            could use instanceof but this is simpler */
        public int mType;

        public static final int TOP         = 0;
        public static final int HASH        = 1;
        public static final int HASHITEM    = 2;
        public static final int HASHITEMKEY = 3;
        public static final int HASHITEMVAL = 4;
        public static final int VECT        = 5;
        public static final int STRING      = 6;
        public static final int LONG        = 7;
        public static final int REFERENCE   = 8;
    };

    /** the content constructor stack: the top of the stack is always
        the object currently being constructed */
    protected Stack mContent;

    protected InProgress peek() 
    {
        return ((InProgress)mContent.peek());
    }

    Object mResult;
    void receive (InProgress ip)
    {
        mResult = ip.mObj;
    }

    // --------------------------------------------------
    // SAX Handlers
    // these are SAX 1.0 handlers, and are deprecated.
    // bite me.

    public void characters (char[] ch, int start, int length)
	throws SAXException
    {
        try {
            peek().characters (ch, start, length);
        } catch (XMLUtilException e) {
            throw new SAXException (e);
        }
    }

    public void ignorableWhitespace (char[] ch, int start, int length)
	throws SAXException
    {
        try {
            peek().characters (ch, start, length);
        } catch (XMLUtilException e) {
            throw new SAXException (e);
        }
    }

    public void startElement (String uri,
                              String name, 
                              String qname,
                              Attributes attributes)
	throws SAXException
    {
        if (!uri.equals (""))
        {
            throw new SAXException ("XMLUtil doesn't use namespaces!");
        }

        try {
            String idatt = attributes.getValue ("id");
            String customatt = attributes.getValue ("custom");
            int type;
            if (name.equals("hash")) type = InProgress.HASH;
            else if (name.equals ("item")) type = InProgress.HASHITEM;
            else if (name.equals ("key")) type = InProgress.HASHITEMKEY;
            else if (name.equals ("value")) type = InProgress.HASHITEMVAL;
            else if (name.equals ("string")) type = InProgress.STRING;
            else if (name.equals ("long")) type = InProgress.LONG;
            else if (name.equals ("vect")) type = InProgress.VECT;
            else if (name.equals ("ref")) type = InProgress.REFERENCE;
            else throw new XMLUtilException ("Unrecognized element: "+name);
            InProgress ip = new InProgress (idatt, customatt, type);
            mContent.push (ip);

            // check for shorthand key attribute and push the key and
            // start a value if so.
            if (type == InProgress.HASHITEM) {
                String keyatt = attributes.getValue ("key");
                if (keyatt != null) {
                    InProgress keyip = 
                      new InProgress (null, null, InProgress.HASHITEMKEY);
                    InProgress keyvalip = 
                      new InProgress (keyatt);
                    keyip.receive (keyvalip);
                    ip.receive (keyip);
                    InProgress valip = 
                      new InProgress (null, null, InProgress.HASHITEMVAL);
                    mContent.push (valip);
                }
            }

            //System.err.println ("start el peek "+peek());
        } catch (XMLUtilException e) {
            throw new SAXException (e);
        }
    }

    public void endElement (String uri,
                            String name,
                            String qname) throws SAXException
    {
        //System.err.println ("end el" + name);
        try {
            // Check for inlined key form
            if (name.equals("item")) {
                if (peek().mType == InProgress.HASHITEMVAL) {
                    InProgress val = ((InProgress)mContent.pop());
                    peek().receive (val);
                }
            }
	    
            InProgress val = (InProgress)mContent.pop ();
            val.filter (this);
            peek().receive (val);
        } catch (XMLUtilException xe) {
            throw new SAXException (xe);
        }
    }

};
